@@ -1,36 +1,50 @@
-/* GLROUTINES.C - general library routines
- * Copyright 1998-2002 Stephen C. Grubb  (ploticus.sourceforge.net) .
- * This code is covered under the GNU General Public License (GPL);
- * see the file ./Copyright for details. */
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
 
-/* OS-specific #defines in rand() below. */
 
 #include <stdio.h> 
+#include <string.h>
 #include <ctype.h>
 #include <time.h>
 #include <stdlib.h>
+
+extern int GL_member(), GL_encode(), GL_wildcmp();
+extern int getpid();
+
 #define stricmp( s, t ) 	strcasecmp( s, t )
 #define strnicmp( s, t, n )     strncasecmp( s, t, n )
 
 #define DATAMAXLEN 256
-#define NUMBER 0
-#define ALPHA 1
 
-static int encode();
-static int getdecplaces();
-static int wcmp();
+static int getdecplaces(), wcmp();
 
 static char Sep = ',';  /* separator character for lists */
-static int domember();
-static char     Gettok_buf[256];
-static char starsym = '*';
+static char Gettok_buf[260];
+static char Wildcard = '*';
+static char Wildcard1 = '?';
 static int Maxlen = 99999;
+static char Member_nullstring[10] = "";
 
+/* ================================= */
+int
+GL_initstatic()
+{
+Sep = ',';
+Wildcard = '*';
+Wildcard1 = '?';
+Maxlen = 99999;
+strcpy( Member_nullstring, "" );
+return( 0 );
+}
 
 /* ================================= */
 /* thanks to Markus Hoenicka for this more portable sysdate and systime code */
 /* SYSDATE - get today's date */ 
 
+int
 GL_sysdate( mon, day, yr )
 int	*mon, *day, *yr ;
 {
@@ -47,6 +61,7 @@ return( 0 );
 
 /* ================================= */
 /* SYSTIME - get current time */ 
+int
 GL_systime( hour, min, sec )
 int	*hour, *min, *sec ;
 {
@@ -61,26 +76,11 @@ return( 0 );
 }
 
 
-/* ===================================================== */
-/* RAND returns a "random" number between 0.0 and 1.0 */
-double GL_rand()
-{
-double r;
-static int first = 1;
-if( first ) {
-	srand( getpid() % 1000 );
-	first = 0;
-	}
-r = rand() / (double)(RAND_MAX);
-if( r < 0.0 || r > 1.0 ) { printf( "%f: rand return out of range\n", r ); exit(1); }
-return( r );
-}
-
-
 /* ============================================= */
-/* GETOK - copied so that buffer size could be increased.. */
+/* GETOK - get next whitespace-delimited token */
  
-char *GL_getok( string, index )
+char 
+*GL_getok( string, index )
 char    *string;
 int     *index;
 {
@@ -91,16 +91,30 @@ for( n=0;
         string[*index] != ' '  &&
         string[*index] != '\t'  &&
         string[*index] != '\n'  &&
+        string[*index] != 13  &&       /* DOS LF added may 03 scg */
         string[*index] != '\0'  ;
                 Gettok_buf[n++] = string[(*index)++] )  ;
 Gettok_buf[n] = '\0' ;
 return( Gettok_buf );
 }
 
+/* ====================================================================== */
+/* MEMBER - returns char position if character c is a member of string s, 
+		0 otherwise. Char positions start with 1 for this purpose.  */
+int
+GL_member( c, s )
+char c, s[];
+{
+int i, len;
+for( i = 0, len = strlen( s ); i < len; i++ ) if( s[i] == c ) return( i+1 );
+return( 0 );
+}
+
 /* ===================================================================== */
 /* SMEMBER - look for s in list t (white-space delimited).  Case sensitive.
    If found return 1 else 0. */
 
+int
 GL_smember( s, t )
 char s[], t[];
 {
@@ -110,6 +124,9 @@ i = 0;
 while( 1 ) {
 	strcpy( tok, GL_getok( t, &i ) );
 	if( tok[0] == '\0' ) break;
+	if( Member_nullstring[0] != '\0' ) {
+		if( strcmp( s, Member_nullstring)== 0 && stricmp( tok, "null" )==0 )return( 1 );
+		}
 	if( strcmp( tok, s ) == 0 ) return( 1 );
 	}
 return( 0 );
@@ -118,6 +135,7 @@ return( 0 );
 /* SMEMBERI - look for s in list t (white-space delimited).  Case insensitive.
    If found return 1 else 0. */
 
+int
 GL_smemberi( s, t )
 char s[], t[];
 {
@@ -127,43 +145,25 @@ i = 0;
 while( 1 ) {
 	strcpy( tok, GL_getok( t, &i ) );
 	if( tok[0] == '\0' ) break;
+	if( Member_nullstring[0] != '\0' ) {
+		if( strcmp( s, Member_nullstring)== 0 && stricmp( tok, "null" )==0 )return( 1 );
+		}
 	if( stricmp( tok, s ) == 0 ) return( 1 );
 	}
 return( 0 );
 }
 
-/* =========================================== */
-/* SLMEMBER - Return 1 if str is matches any items in list.
-	List is a space-delimited list of tokens.  The tokens in the
-	list may contain ? or * wildcard characters.  The match is
-	Case insensitive.
 
-	scg 3-19-97
-*/
-GL_slmember( str, list )
-char *str;
-char *list;
+
+/* ==================================================================== */
+/* MEMBER_NULLMODE - set a special mode for the benefit of shsql, to consider
+   "null" or "NULL" equivalent to some code such as "=" */
+
+int
+GL_member_nullmode( s )
+char *s;
 {
-char tok[100], *GL_getok();
-int i;
-i = 0;
-while( 1 ) {
-        strcpy( tok, GL_getok( list, &i ) );
-        if( tok[0] == '\0' ) break;
-        if( GL_wildcmp( str, tok, strlen(tok), 0 ) == 0 ) return( 1 );
-        }
-return( 0 );
-}
-
-
-/* ====================================================================== */
-/* MEMBER - returns char position if character c is a member of string s, 
-		0 otherwise. Char positions start with 1 for this purpose.  */
-GL_member( c, s )
-char c, s[];
-{
-int i, len;
-for( i = 0, len = strlen( s ); i < len; i++ ) if( s[i] == c ) return( i+1 );
+strcpy( Member_nullstring, s );
 return( 0 );
 }
 
@@ -173,10 +173,11 @@ return( 0 );
 	'prec' is returned.. and is the precision of the number (position of decimal point).
 
 	Number may contain unary + or -, and one decimal point.
-	Leading and trailing whitespace are discarded before determining
-	whether the case is a valid number or not.
+
+	Leading and trailing whitespace are tolerated by this routine, eg. "  28.5 " is considered OK
 */
 
+int
 GL_goodnum( str, prec )
 char *str;
 int *prec;
@@ -187,11 +188,11 @@ bad = 0; *prec = -1;
 len = strlen( str );
 
 /* find limit of trailing whitespace.. */
-for( i = len-1; i >= 0; i-- ) if( !isspace( str[i] ) ) break;
+for( i = len-1; i >= 0; i-- ) if( !isspace( (int) str[i] ) ) break;
 len = i+1;
 
 /* skip over leading whitespace.. */
-for( i = 0; i < len; i++ ) if( !isspace( str[i] ) ) break;
+for( i = 0; i < len; i++ ) if( !isspace( (int) str[i] ) ) break;
 start = i;
 
 /* screen out degenerate cases.. */
@@ -206,7 +207,12 @@ for( p = start; p < len; p++ ) {
 		else bad=1; 
 		}
 	else if( p == start && ( str[p] == '-' || str[p] == '+' ) );
-	else if( ! isdigit( str[p]) ) bad=1;
+	else if( p > start && tolower(  (int) str[p]) == 'e' ) {  /* handle scientific notation ... scg 7/29/04 */
+		p++;
+		if( str[p] != '-' && str[p] != '+' ) bad=1;
+		if( p+1 >= len ) bad=1;
+		}
+	else if( ! isdigit( (int) str[p]) ) bad=1;
 	}
 
 /* return.. */
@@ -226,7 +232,7 @@ else return( 1 );
    SCG 12-2-96
 */
 
-
+int
 GL_getseg( rtn, inbuf, i, sep )
 char rtn[];
 char inbuf[];
@@ -263,6 +269,7 @@ return( eol );
 /* ==================================================================== */
 /* GETCHUNK - Get tokens, which are separated by any member of sepstring */
 
+int
 GL_getchunk( rtn, line, i, sepstring )
 char rtn[];
 char line[];
@@ -281,11 +288,166 @@ while( 1 ){
 	if( n >= (Maxlen-1) ) break;
 	}
 rtn[n] = '\0' ;
+return( 0 );
 }
 
+/* ============================================= */
+/* MAKE_UNIQUE_STRING - generate an identifier using date, time, and pid */
+
+int
+GL_make_unique_string( s, i )
+char *s;
+int i;  /* may be sent as an integer.. if 0 getpid() will be used.. */
+{
+int mon, day, yr, hr, min, sec, pid;
+GL_sysdate( &mon, &day, &yr );
+GL_systime( &hr, &min, &sec );
+s[0] = GL_encode( yr % 100 );
+s[1] = GL_encode( mon );
+s[2] = GL_encode( day );
+s[3] = GL_encode( hr );
+s[4] = GL_encode( min );
+s[5] = GL_encode( sec );
+if( i == 0 ) pid = getpid();
+else pid = i;
+s[6] = GL_encode( pid % 62 );
+pid = pid / 62;
+s[7] = GL_encode( pid % 62 );
+s[8] = GL_encode( pid / 62 );
+s[9] = '\0';
+
+return( 0 );
+}
+
+/* encode - derive a character representation of a number (number must be in range 0 - 62) */
+int
+GL_encode( a )
+int a;
+{
+if( a >= 0 && a <= 9 ) return( a + '0' );
+else if( a > 35 ) return( (a-36) + 'A' ); /* A-Z    26 letters + 9 = 35 */
+else if( a > 9 ) return( (a-10) + 'a' ); /* a-z */
+else return( '0' );
+}
+
+/* decode - decode a character representation of a number */
+int
+GL_decode( a )
+int a;
+{
+if( a >= '0' && a <= '9' ) return( a - '0' );
+else if( a >= 'a' ) return( (a - 'a')+10 ); /* a-z */
+else if( a >= 'A' ) return( (a - 'A')+36 ); /* A-Z    26 letters + 9 = 35 */
+else return( '0' );
+}
+
+/* ============================================= */
+/* EXPAND_TABS Takes a string parameter 'in' and expands tabs into spaces, placing the
+      result into parameter 'out'.  
+ */
+int
+GL_expand_tabs( out, in )
+char in[], out[];
+{
+int i, j, k, l, len;
+
+out[0] = '\0';
+k = 0;
+for( i = 0, len = strlen( in ); i < len; i++ ) {
+	if( in[i] == '\t' ) {
+		j =  8 - ( k % 8 ); /* 1 to 8 spaces needed */
+		for( l = 0; l < j; l++ ) out[k++] = ' ';
+		}
+	else out[k++] = in[i];
+	}
+out[k] = '\0';
+return( 0 );
+}
+
+
+/* ==================================================== */
+/* WRAPTEXT - wrap txt so that no line exceeds maxchars.  Wrap is done by changing certain whitespace chars to '\n' */
+int GL_wraptext( txt, maxchars )
+char *txt;    /* the text */
+int maxchars; /* max # of chars per line after wrap */
+{
+int i, lb, spaceat;
+lb = 0;
+spaceat = -1;
+for( i = 0; txt[i] != '\0'; i++ ) {
+        if( i - lb > maxchars ) {
+                if( spaceat == -1 ) spaceat = i; /* for wierd situations - no space found, break right here */
+                txt[spaceat] = '\n';
+                lb = spaceat;
+                spaceat = -1;
+                }
+        else if( txt[i] == '\n' ) { lb = i; spaceat = -1; } /* newline already present in txt.. respect it.. */
+        else if( isspace( (int) txt[i] ) ) spaceat = i; 
+        }
+return( 0 );
+}
+
+/* test for wraptext */
+/* main( argc, argv )
+ * int argc; char **argv;
+ * {
+ * char buf[256];
+ * if( argc < 2 ) exit(1);
+ * strcpy( buf, argv[1] );
+ * GL_wraptext( buf, atoi( argv[2] ) );
+ * printf( "%s\n", buf );
+ * }
+ */
+
+
+#ifndef BAREBONES  /* getgui, lxlogo */
+
+/* ===================================================== */
+/* RAND returns a "random" number between 0.0 and 1.0 */
+double 
+GL_rand()
+{
+double r;
+static int first = 1;
+if( first ) {
+	srand( getpid() % 1000 );
+	first = 0;
+	}
+r = rand() / (double)(RAND_MAX);
+if( r < 0.0 || r > 1.0 ) { printf( "%f: rand return out of range\n", r ); return( -1.0 ); }
+return( r );
+}
+
+
+/* =========================================== */
+/* SLMEMBER - Return 1 if str is matches any items in list.
+	List is a space-delimited list of tokens.  The tokens in the
+	list may contain ? or * wildcard characters.  The match is
+	Case insensitive.
+
+	scg 3-19-97
+*/
+int
+GL_slmember( str, list )
+char *str;
+char *list;
+{
+char tok[256], *GL_getok();
+int i;
+i = 0;
+while( 1 ) {
+        strcpy( tok, GL_getok( list, &i ) );
+        if( tok[0] == '\0' ) break;
+	if( Member_nullstring[0] != '\0' ) {
+		if( strcmp( str, Member_nullstring)== 0 && stricmp( tok, "null" )==0 )return( 1 );
+		}
+        if( GL_wildcmp( str, tok, strlen(tok), 0 ) == 0 ) return( 1 );
+        }
+return( 0 );
+}
 /* ==================================================================== */
 /* SETMAXLEN - set maximum token length for GETSEG (future: others) */
-
+int
 GL_setmaxlen( maxlen )
 int maxlen;
 {
@@ -312,7 +474,7 @@ return( 0 );
 
  */
 
-
+int
 GL_wildcmp( char *s1, char *s2, int len, int casecare )
 /* s1  = data value
    s2  = query value which can contain wildcards - not null terminated.
@@ -320,24 +482,24 @@ GL_wildcmp( char *s1, char *s2, int len, int casecare )
    casecare = 0 for case-insensitive, 1 for case-sensitive
  */
 {
-int i, nwc, wcp, stat, stat2;
+int i, nwc, wcp, stat;
 
 
 if( len == 0 ) return( strlen( s1 ) );
-else if( s2[0] == starsym ) {
+else if( s2[0] == Wildcard ) {
 	if( len == 1 ) return( 0 ); /* everything matches */
 	}
-else if( s2[0] == '?' ) ; /* can't tell yet */
+else if( s2[0] == Wildcard1 ) ; /* can't tell yet */
 else if( tolower( s1[0] ) < tolower( s2[0] ) ) return( -1 ); /* way off */
 else if( tolower( s1[0] ) > tolower( s2[0] ) ) return( 1 );  /* way off */
 
 /* strip off extraneous * at beginning and end.. */
-if( s2[0] == starsym && s2[1] == starsym ) { s2 = &s2[1]; len--; }
-if( s2[len-1] == starsym && s2[len-2] == starsym ) len--;
+if( s2[0] == Wildcard && s2[1] == Wildcard ) { s2 = &s2[1]; len--; }
+if( s2[len-1] == Wildcard && s2[len-2] == Wildcard ) len--;
 
-/* see if any wild cards were used.. */
+/* see if any "*" wild cards were used.. */
 nwc = 0;
-for( i = 0; i < len; i++ ) if( s2[i] == starsym ) { nwc++; wcp = i; }
+for( i = 0; i < len; i++ ) if( s2[i] == Wildcard ) { nwc++; wcp = i; }
 
 if( nwc < 1 ) {  /* straight match */
 	if( strlen( s1 ) > len ) return( wcmp( s1, s2, strlen( s1 ), casecare));
@@ -382,7 +544,7 @@ else if( nwc == 2 ) {
 	int stop;
 	/* case 4: wc at beginning and end.. */
 	if( wcp != (len-1) ) goto ERR;
-	else if( s2[0] != starsym ) goto ERR;
+	else if( s2[0] != Wildcard ) goto ERR;
 	stop = ( strlen( s1 ) - len ) + 2;
 	for( i = 0; i <= stop; i++ ) {
 		if( wcmp( &s1[i], &s2[1], len-2, casecare ) == 0 ) return( 0 );
@@ -404,198 +566,45 @@ static int
 wcmp( char *s1, char *s2, int len, int casecare )
 {
 int i;
+
 for( i = 0; i < len; i++ ) {
+	
+	if( s1[i] == '\0' ) { return( -1 ); }  /* added scg 10/22/03 ... abc???? was matching abcde */
+									   /* was returning 1.. changed scg 3/29/04 */
+
 	if( ! casecare ) {
-		if( tolower(s1[i]) < tolower(s2[i]) && s2[i] != '?' ) 
-			return( -1 );
-		else if( tolower(s1[i]) > tolower(s2[i]) && s2[i] != '?' ) 
-			return( 1 );
+		if( tolower(s1[i]) < tolower(s2[i]) && s2[i] != Wildcard1 ) { return( -1 ); }
+		else if( tolower(s1[i]) > tolower(s2[i]) && s2[i] != Wildcard1 ) { return( 1 ); }
 		}
 	else	{
-		if( s1[i] < s2[i] && s2[i] != '?' ) return( -1 );
-		else if( s1[i] > s2[i] && s2[i] != '?' ) return( 1 );
+		if( s1[i] < s2[i] && s2[i] != Wildcard1 ) return( -1 );
+		else if( s1[i] > s2[i] && s2[i] != Wildcard1 ) return( 1 );
 		}
 	}
 return( 0 );
 }
 
 /* WILDCHAR - set the wildcard symbol to be used instead of '*' */
-GL_wildchar( c )
-char c;
+int
+GL_wildchar( c, d )
+char c, d;
 {
-starsym = c;
+Wildcard = c;
+Wildcard1 = d;
 return( 0 );
 }
 
-/* ===================================================================== */
-/* FUZZYMATCH - Do a 'fuzzy' match.  See if s1 is similar to s2.
 
-   This routine may alter s1 and s2.
-
-   A wildcard character '*' at the beginning or end of s2 indicates
-   that a match can occur anywhere within s1 (eg. smith* would match
-   smithington and *smith would match harrowsmith; *smith* would match
-   both).  If there are no asterisks, lengths
-   of the two strings must be similar.
-
-   Degree changes the 'looseness' of the match.  
-   5 = strict, 4 = medium-strict, 3 = medium 2= loose, 1 = very loose.
-   
-   Returns 1 on a match, 0 not. 
-
-   SCG '94,'95
- */
-
-GL_fuzzymatch( s1, s2, len2, degree )
-char *s1; /* data value */
-char *s2; /* query value */
-int len2; /* length of s2 */
-int degree; /* dgree of tightness */
-{
-int i, j, k, len1, pts;
-int bestpts, score;
-char c;
-int openfront, openend, goal;
-
-/* exact match.. no need to do more.. */
-if( strcmp( s2, s1 ) ==0 ) return( 1 );
-
-len1 = strlen( s1 );
-
-/* see what type of match, open or closed */
-if( s2[0] == '*' ) { openfront = 1; s2 = &s2[1]; len2--; }
-else openfront = 0;
-if( s2[ len2-1 ] == '*' ) { openend = 1; len2--; }
-else openend = 0;
-
-/* set goal */
-if( len2 < 4 )goal = len2 * degree;
-else if( len2 < 10 ) goal = (len2 -2) * degree;
-else goal = (len2 -3) * degree;
-
-
-/* reject disparate lengths if doing a closed match */
-if( !openfront && !openend && len1 - len2 > 3 ) return( 0 );
-
-/* truncate end of s1 if openended match */
-if( openend && !openfront && len1 > len2+2 ) len1 = len2+2;
-
-/* chop of front of s1 if openfronted match */
-if( openfront && !openend && len1 > len2+2 ) s1 = &s1[ len1-(len2+2) ];
-
-/* printf( "[%d-%d]", openfront, openend ); */
-
-/* go thru s2.. */
-score = 0;
-for( i = 0; i < len2; i++ ) {
-	c = s2[i];
-
-	/* find each occurrence of c in s1.. */
-	bestpts = 0;
-	pts = 0;
-	for( j = 0; j < len1; j++ ) {
-
-		if( s1[j] == c ) {
-
-			/* check adjacent characters, count matches.. */
-			pts = 1;
-
-			if( i + j == 0 ) pts+=2; /* both 1st char */
-			
-			else /* go left.. */
-			  for( k = 1; k <= 4; k++ ) {
-				if( j-k < 0 || i-k < 0 ) break;
-				else if( s1[j-k] == s2[i-k] ) pts++;
-				else break;
-				}
-			if( i == len2-1 && j == len1-1 )
-				pts+=2; /* both last char */
-
-			else /* go right.. */
-			  for( k = 1; k <= 4; k++ ) {
-				if( j+k > len1-1 || i+k > len2-1 ) break;
-				else if( s1[j+k] == s2[i+k] ) pts++;
-				else break;
-				}
-			}
-		if( pts > bestpts )bestpts = pts;
-		}
-	if( bestpts > 6 ) score += 6;
-	else score += bestpts;
-	/* printf( "[%c:%d]", c, bestpts ); */
-	if( score > goal ) return( 1 ); /* optimize */
-	}
-/* printf( "<%s v. %s : %d>\n", s2, s1, score ); */
-if( score > goal ) return( 1 );
-else return( 0 );
-}
-
-
-/* ============================================= */
-/* EXPAND_TABS Takes a string parameter 'in' and expands tabs into spaces, placing the
-      result into parameter 'out'.  
- */
-GL_expand_tabs( out, in )
-char in[], out[];
-{
-int i, j, k, l, len;
-
-out[0] = '\0';
-k = 0;
-for( i = 0, len = strlen( in ); i < len; i++ ) {
-	if( in[i] == '\t' ) {
-		j =  8 - ( k % 8 ); /* 1 to 8 spaces needed */
-		for( l = 0; l < j; l++ ) out[k++] = ' ';
-		}
-	else out[k++] = in[i];
-	}
-out[k] = '\0';
-}
-
-
-/* ============================================= */
-/* MAKE_UNIQUE_STRING - generate an identifier using date, time, and pid */
-
-GL_make_unique_string( s, i )
-char *s;
-int i;  /* may be sent as an integer.. if 0 getpid() will be used.. */
-{
-int mon, day, yr, hr, min, sec, pid, a, b, c;
-GL_sysdate( &mon, &day, &yr );
-GL_systime( &hr, &min, &sec );
-s[0] = encode( yr % 100 );
-s[1] = encode( mon );
-s[2] = encode( day );
-s[3] = encode( hr );
-s[4] = encode( min );
-s[5] = encode( sec );
-if( i == 0 ) pid = getpid();
-else pid = i;
-s[6] = encode( pid % 62 );
-pid = pid / 62;
-s[7] = encode( pid % 62 );
-s[8] = encode( pid / 62 );
-s[9] = '\0';
-
-return( 0 );
-}
-
-/* encode - derive a character representation of a number */
-static int encode( a )
-int a;
-{
-if( a >= 0 && a <= 9 ) return( a + '0' );
-else if( a > 35 ) return( (a-36) + 'A' ); /* A-Z    26 letters + 9 = 35 */
-else if( a > 9 ) return( (a-10) + 'a' ); /* a-z */
-else return( '0' );
-}
 
 /* ============================================= */
 /* ADDMEMBER - append a new member to the end of a comma-delimited list */
+int
 GL_addmember( newmem, list )
 char *newmem;
 char *list;
 {
+/* hard code a destination length limit of 256 for now.. scg 3/22/07 */
+if( strlen( newmem ) + strlen( list ) > 254 ) return( 0 );  /* silently don't do it.. */
 if( list[0] == '\0' ) strcpy( list, newmem );
 else	{
 	strcat( list, "," );
@@ -604,11 +613,43 @@ else	{
 return( 0 );
 }
 
+/* ============================================= */
+/* DELETEMEMBER - remove member(s) from a comma-delimited list.
+   Mem may contain wild cards.  Returns number of members removed. */
+int
+GL_deletemember( mem, inlist, resultlist )
+char *mem;
+char *inlist;
+char *resultlist;
+{
+int i, ix, len, outlen, found, memlen;
+char tok[ 256 ];
+
+resultlist[0] = '\0';
+outlen = 0;
+if( inlist[0] == '\0' ) return( 1 );
+memlen = strlen( mem );
+len = strlen( inlist );
+for( i = 0, ix = 0, found = 0; ; ) {
+        if( ix >= len ) break;
+        GL_getseg( tok, inlist, &ix, "," );
+        if( GL_wildcmp( tok, mem, memlen, 0 )==0 ) { found++; continue; }
+	if( i > 0 ) strcpy( &resultlist[ outlen++ ], "," );
+	strcpy( &resultlist[ outlen ], tok );
+	outlen += strlen( tok );
+	i++;
+	}
+resultlist[ outlen ] = '\0';
+return( found );
+}
+
+
 /* ============================================= */
 /* CONTAINS - if string s contains any of chars in clist, return position (1=first)
      of first occurance in list.  0 if not found at all.
      example: contains( "\"*'", "'hello'" )  -> 1
  */
+int
 GL_contains( clist, s )
 char *clist, *s;
 {
@@ -623,33 +664,36 @@ return( 0 );
 /* ============================================= */
 /* SUBST - change all occurances of s1 to s2, in t.
 
-   Max length of t is 255.
-   Returns 0 if successful, 1 if no occurance of s1 found. */
+   Returns 0 if successful, 1 if no occurance of s1 found. 
+   scg 11/5/07 - now silently truncates the result to fit within buf[1024]
+*/
 
+int
 GL_substitute( s1, s2, t )
 char *s1, *s2, *t;
 {
-char buf[255];
-int i, j, len1, buflen, found;
+char buf[1024];
+int i, j, len1, len2, buflen, found;
 
 len1 = strlen( s1 );
 if( len1 < 1 ) return( 1 );
 strcpy( buf, t );
 buflen = strlen( buf );
 if( buflen < 1 ) return( 1 );
+len2 = strlen( s2 );
 
 found = 0;
 j = 0;
 for( i = 0; i < buflen; i++ ) {
-/* 	printf( "[%s|%s|%d]", &buf[i], s1, len1 ); */
-
 	if( strncmp( &buf[i], s1, len1 )==0 ) {
+		if( j + len2 > 1020 ) break; /* scg 11/3/07 */
 		strcpy( &t[j], s2 );
-		j += strlen( s2 );
+		j += len2;
 		i += (len1 - 1);
 		found = 1;
 		}
 	else t[j++] = buf[i];
+	if( j > 1020 ) break; /* scg 11/3/07 */
 	}
 t[j] = '\0';
 if( found ) return( 0 );
@@ -659,14 +703,20 @@ else return( 1 );
 /* ============================================= */
 /* CHANGECHARS - go through string s and if any characters in clist found, change
 	the character to newchar */
+
+int
 GL_changechars( clist, s, newchar )
 char *clist, *s, *newchar;
 {
+int i, special;
 
-int i, len;
+special = 0;
+if( strcmp( clist, "not_alnum" )==0 ) special = 1;
 
-for( i = 0, len = strlen( s ); i < len; i++ ) {
-	if( GL_member( s[i], clist )) s[i] = newchar[0];
+if( newchar[0] == '\0' ) return( 1 );
+for( i = 0; s[i] != '\0'; i++ ) {
+	if( special && !isalnum( (int)s[i] )) s[i] = newchar[0];
+	else if( !special && GL_member( s[i], clist )) s[i] = newchar[0];
 	}
 return( 0 );
 }
@@ -674,20 +724,22 @@ return( 0 );
 /* ============================================= */
 /* DELETECHARS - go through string s and if any characters in clist found, delete
 	the character. */
+
+int
 GL_deletechars( clist, s )
 char *clist, *s;
 {
+int i, j, special;
 
-int i;
+special = 0;
+if( strcmp( clist, "not_alnum" )==0 ) special = 1;
 
-for( i = 0; ; ) {
-	if( s[i] == '\0' ) break;
-	if( GL_member( s[i], clist )) {
-		strcpy( &s[i], &s[i+1] );
-		continue;
-		}
-	else i++;
+for( i = 0, j = 0; s[i] != '\0';  ) {
+	if( special && !isalnum( (int)s[i] )) i++; 
+	else if( !special && GL_member( s[i], clist )) i++; 
+	else s[j++] = s[i++];
 	}
+s[j] = '\0';
 return( 0 );
 }
 /* ======================================================================== */
@@ -714,6 +766,7 @@ return( 0 );
               substring( result, "02001.fv02", -4, 99 )   -- result would be "fv02"
 */
 
+int
 GL_substring( result, str, fromchar, nchar )
 char *result;
 char *str;
@@ -738,29 +791,36 @@ return( 0 );
 /* ===================================================================== */
 /* VARSUB - given string s, find every occurance 
    of symbol (case-sensitive) and change it to value.  
-    -Copies result into s.  
-     -Returns number of times a substitution was made, 0 if none.
-     -This routine is not sophisticated about delimiting the symbol;
+    Copies result back into s... s must be able to accomodate.
+    Returns number of times a substitution was made, 0 if none.
+    This routine is not sophisticated about delimiting the symbol;
       e.g. if s contains $NUMBER and varsub() is looking for $NUM it will find it.
+
+    -scg 11/5/07 - now silently truncates result to fit within rtnbuf[1024]
 */
+
+int
 GL_varsub( s, symbol, value )
 char *s, *symbol, *value;
 {
-int i, j, len, found;
+int i, j, len, vlen, found;
 int slen;
-char rtnbuf[256];
+char rtnbuf[1024];
 
 len = strlen( symbol );
 slen = strlen( s );
+vlen = strlen( value );
 found = 0;
-for( i = 0, j = 0; i < slen; i++, j++ ) {
-	if( strncmp( &s[i], symbol, len )==0 ) {
+for( i = 0, j = 0; i < slen; i++, j++ ) { 
+	if( strncmp( &s[i], symbol, len )==0 ) {  /* note- strncmp man page says that it won't go beyond null terminator */
+		if( j + vlen > 1020 ) break;
 		strcpy( &rtnbuf[j], value );
 		j = strlen( rtnbuf ) - 1;
 		i+= (len-1);
 		found++;
 		}
 	else rtnbuf[j] = s[i]; 
+	if( j >= 1020 ) break; /* scg 11/3/07 */
 	}
 rtnbuf[j] = '\0';
 strcpy( s, rtnbuf );
@@ -781,6 +841,8 @@ return( found );
 
    If val is non-numeric or a whole number then it is left unchanged.
 */
+
+int
 GL_autoround( val, decoffset )
 char *val;
 int decoffset; 
@@ -804,12 +866,13 @@ return( 0 );
 
 /* ============================================= */
 /* AUTOROUNDF - variant of autoround(), takes val as a double, return value is character rep..*/
+
 char *
 GL_autoroundf( val, decoffset )
 double val;
 int decoffset;
 {
-int precision, decplaces, stat;
+int decplaces;
 char roundingfmt[50];
 static char result[50];
 
@@ -843,6 +906,7 @@ else if( g >= 0.0001 ) decplaces = 6;
 return( decplaces );
 }
 
+#ifdef HOLD
 /* ======================== */
 /* FMOD  */
 double fmod( a, b )
@@ -854,6 +918,29 @@ x = a / b;
 y = (int) (x) * b;
 return( a - y );
 }
+#endif
+
+/* ========================= */
+/* NUMGROUP - convert val to a nearby multiple of h, taking mode (low, mid, high) into account */
+
+double
+GL_numgroup( val, h, mode )
+double val, h;
+char *mode;
+{
+double fmod(), ofs, modf;
+ofs = 0.0;
+if( mode[0] == 'm' ) ofs = h / 2.0;
+else if( mode[0] == 'h' ) ofs = h;
+/* the following bug fix contributed by  Phil Carmody <thefatphil@yahoo.co.uk> */
+modf = h*(int)(val/h+0.5);
+return(modf+ofs);
+/* was:
+ * modf = fmod( val, h );
+ * return( (val - modf) + ofs );
+ */
+
+}
 
 
 /* ======================================================================== */
@@ -866,6 +953,7 @@ return( a - y );
  * 	    There may be no embedded spaces within the dash construct.
  */
 
+int
 GL_ranger( spec, list, n )
 char *spec;
 int *list;  /* array */
@@ -873,7 +961,7 @@ int *n;  /* in: size of list (max number of members)
 	    out: number of members in list that have been filled */
 {
 int i, ix, p, j, lo, hi;
-char tok[80], histr[80];
+char tok[256], histr[80];
 
 /* parse spec.. */
 ix = 0;
@@ -898,6 +986,7 @@ while( 1 ) {
 		for( j = lo; j <= hi; j++ ) {
 			list[i] = j;
 			i++;
+			if( i >= (*n) -1 ) break;  /* truncate */
 			}
 		}
 	}
@@ -910,6 +999,7 @@ return( 0 );
 /* CLOSE_TO - test two floating point numbers to see if
         they are within a small tolerance. */
 
+int
 GL_close_to( a, b, tol )
 double a, b;
 double tol;
@@ -921,98 +1011,11 @@ else return( 0 );
 }
 
 
-/* ================================= */
-/* CHECKSUM FUNCTIONS  */
-
-/* CHECKDIG - returns the check digit for the number in aptr with length l.
- 	NOTE: when the check digit is computed to be 10 an `x' is returned.  
-
-   This algorithm protects against interchanged digits.
-*/
- 
-#define TEN	'x'
-
-char GL_checkdig(aptr,l)
-char *aptr;
-int l;
-{
-int i,odds,evens,checkdigit;
-	
-odds = evens = 0;
-i = 0;
-
-while (i < l) {
-	if (i % 2 == 0) { evens += (*(aptr+i) - '0'); }
-	else { odds += (*(aptr+i) - '0'); }
-	i++;
-	}
-checkdigit = (odds + (3 * evens)) % 11;
-	
-/* if the checkdigit is 10 then return the character to be used in place of ten */
-if ( checkdigit == 10 ) return ( TEN );
-else return(checkdigit+'0'); 
-}
-
-
-/* ------------ */
-/* script entry point for checksum functions */
-GL_checksum_functions( hash, name, arg, nargs, result, typ )
-int hash;
-char *name;
-char *arg[];
-int nargs;
-char *result;
-int *typ;
-{
-char *s;
-
-if( hash == 2182 ) { /* $checksumvalid(s) -  validate s which is an integer containing a trailing check digit, 
-   			using the algorithm in $ps/clibx/checkdig  */
-	int i;
-	i = GL_checkdig( arg[0], strlen( arg[0] ) -1 );
-	if( i == arg[0][ strlen( arg[0] ) -1 ] ) sprintf( result, "1" );
-	else sprintf( result, "0" );
-	*typ = NUMBER;
-	return( 0 );
-	}
-
-if( hash == 2412 ) { /* $checksumencode(i) - result is i with checksum digit appended. */
-	long i, atol();
-	char tmp[20];
-	int cs;
-	i = atol( arg[0] );
-	sprintf( tmp, "%ld", i );
-	cs = GL_checkdig( tmp, strlen( tmp ) );
-	sprintf( result, "%s%c", tmp, cs );
-	*typ = ALPHA;
-	return( 0 );
-	}
-	
-if( hash == 2155 ) { /* $checksumnext(s) take s which is a number including trailing checksum 
-   			digit, increment number and recompute new checksum digit. */
-	int cs;
-	long i, atol();
-	char tmp[20];
-	s = arg[0];
-	s[ strlen( s ) - 1 ] = '\0'; /* strip off existing checksum digit */
-	i = atol( s );
-	i++;
-	sprintf( tmp, "%ld", i );
-	cs = GL_checkdig( tmp, strlen( tmp ) );
-	sprintf( result, "%s%c", tmp, cs );
-	*typ = ALPHA;
-	return( 0 );
-	}
-
-fprintf( stderr, "unrecognized function: %s\n", name ); /* not found */
-return( 195 );
-}
-
-
 /* ============================= */
 /* COMMONMEMBERS - compare two commalists and return number of members
    that are in common.. */
 
+int
 GL_commonmembers( list1, list2, mode )
 char *list1;
 char *list2;
@@ -1043,6 +1046,8 @@ return( count );
 /* ==================================== */
 /* LISTMEMBER - see if s is in list (comma-delimited); 
    if so return 1 and list position (first=1) and string position (first=0) */
+
+int
 GL_listmember( s, list, memnum, pos )
 char *s;
 char *list;
@@ -1070,7 +1075,9 @@ return( 0 );
 }
 
 /* ===================================== */
-/* GETCGIARG - get next arg from CGI REQUEST_URI string (escape constructs are converted) */
+/* GETCGIARG - get next arg from CGI QUERY_STRING (encoded constructs are converted) */
+
+int
 GL_getcgiarg( arg, uri, pos, maxlen )
 char *arg, *uri;
 int *pos; /* current position */
@@ -1080,33 +1087,125 @@ int i, j;
 char hex[10];
 unsigned int val;
 
-if( *pos == 0 ) {   /* scan to '?'.. */
-	for( i = 0; ; i++ ) {
-		if( uri[i] == '?' || uri[i] == '\0' ) {
-			strncpy( arg, uri, i );
-			arg[i] = '\0';
-			if( uri[i] == '\0' ) *pos = i;
-			else *pos = i+1;
-			return( 0 );
-			}
+for( i = *pos, j = 0; j < maxlen; i++ ) {
+
+	if( uri[i] == '&' || uri[i] == '\0' || j >= maxlen ) {
+		arg[j] = '\0';
+		if( uri[i] == '\0' ) *pos = i;
+		else *pos = i+1;
+		return( 0 );
 		}
-	}
-else	{
-	for( i = *pos, j = 0; j < maxlen; i++ ) {
-		if( uri[i] == '&' || uri[i] == '\0' || j >= maxlen ) {
-			arg[j] = '\0';
-			if( uri[i] == '\0' ) *pos = i;
-			else *pos = i+1;
-			return( 0 );
-			}
-		else if( uri[i] == '%' && isxdigit( uri[i+1] ) && isxdigit( uri[i+2] ) ) {
-			sprintf( hex, "%c%c", uri[i+1], uri[i+2] );
-        		sscanf( hex, "%x", &val );
-			arg[j++] = (char) val;
-			i += 2;
-			}
-		else arg[j++] = uri[i];
+
+	else if( uri[i] == '%' && isxdigit( (int) uri[i+1] ) && isxdigit( (int) uri[i+2] ) ) {  /* urldecode */
+		sprintf( hex, "%c%c", uri[i+1], uri[i+2] );
+       		sscanf( hex, "%x", &val );
+		arg[j++] = (char) val;
+		i += 2;
 		}
+
+	else if( uri[i] == '+' ) arg[j++] = ' ';  /* added scg 10/9/07 */
+
+	else arg[j++] = uri[i];
+
 	}
+
 return( 0 );
 }
+
+/* ================================================= */
+/* URLENCODE - perform url encoding (any questionable characters changed to %XX hex equivalent */
+/* added scg 5/29/06 */
+int
+GL_urlencode( in, out )
+char *in, *out;
+{
+int i, j, c;
+for( i = 0, j = 0; in[i] != '\0'; i++ ) {
+	c = in[i];
+
+	/* per the wikipedia entry for "Query string"... changed scg 6/4/07 */
+	if( c >= 48 && c <= 57 ) out[j++] = c;  /* 0-9 */
+	else if( c >= 65 && c <= 90 ) out[j++] = c;  /* A-Z */
+	else if( c >= 97 && c <= 122 ) out[j++] = c;  /* a-z */
+	else if( GL_member( c, ".-~_" )) out[j++] = c; 
+	/* else if( c == ' ' ) out[j++] = '+'; */  /* encode spaces as %20 ... scg 3/9/09 */
+	else { sprintf( &out[j], "%%%X", c ); j += 3; } /* encode as %FF */
+	}
+out[j] = '\0'; /* terminate */
+return( 0 );
+}
+/* ================================================= */
+/* URLDECODE - perform url decoding (any %XX constructs changed to char equivalent */
+/* added scg 5/29/06 */
+int GL_urldecode( in, out )
+char *in, *out;
+{
+int i, j, c;
+char tok[10];
+for( i = 0, j = 0; in[i] != '\0'; i++ ) {
+	if( in[i] == '%' && in[i+1] != '\0' && in[i+2] != '\0' ) {
+		tok[0] = in[i+1]; tok[1] = in[i+2]; tok[2] = '\0';
+		sscanf( tok, "%x", &c );
+		out[j++] = c;
+		i += 2;
+		}
+	else out[j++] = in[i];
+	}
+out[j] = '\0'; /* terminate */
+return( 0 );
+}
+
+
+/* ================================================= */
+/* GETCWORD - get next word, as delimited by any sequence of spaces and punct chars - related to 'contains' */
+
+int
+GL_getcword( rtn, line, i )
+char rtn[];
+char line[];
+int *i;
+{
+int n, j;
+j = *i;
+while( isspace( (int) line[j] ) || ispunct( (int) line[j] ) ) j++; 
+n = 0;
+rtn[0] = '\0';
+while( 1 ){
+	if( line[j] != '*' && ( isspace( (int) line[j] ) || ispunct( (int) line[j] ) || line[j] == '\0' )) break;
+	else rtn[n++] = line[j];
+	j++;
+	}
+*i = j;
+rtn[n] = '\0' ;
+return( 0 );
+}
+
+/* =================================== */
+/* STRIP_WS strip white-space off of front and end of string s */
+
+int
+GL_strip_ws( s )
+char *s;
+{
+int i, j, len;
+
+/* don't do anything if first and last characters are non-space.. */
+if( !isspace( (int) s[0] ) && !isspace( (int) s[ strlen( s ) - 1 ] ) ) return( 0 );
+ 
+/* find last significant char and put a null after it */
+for( j = strlen( s ) -1; j >= 0; j-- )
+	if( !GL_member( s[j], " \t\n" )) break;
+s[j+1] = '\0';
+/* find 1st significant char at position i */
+for( i = 0, len = strlen( s ); i < len; i++ ) 
+	if( !GL_member( s[i], " \t\n" )) break; 
+strcpy( s, &s[i] );
+return( 0 );
+}
+#endif
+
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
