@@ -1,6 +1,8 @@
-/* ploticus data display engine.  Software, documentation, and examples.  
- * Copyright 1998-2002 Stephen C. Grubb  (scg@jax.org).
- * Covered by GPL; see the file ./Copyright for details. */
+/* ======================================================= *
+ * Copyright 1998-2008 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
 
 /* PROC TABULATE - do frequency distributions, 1-D or 2-D */
 
@@ -11,6 +13,8 @@
                   added accumfield option.
    scg 1/21/00  - for 2-way dist, column tags are used to set getdata field names
 
+   scg 4/2/04   - added numfmt - %g doesn't give correct results for sets of very small values
+
 */
 #include "pl.h"
 
@@ -22,62 +26,31 @@
 
 static int fsort(), freqsort();
 
+int
 PLP_tabulate()
 {
-int i;
-char attr[40], val[256];
-char *line, *lineval;
-int nt, lvp;
-int first;
+char attr[NAMEMAXLEN], *line, *lineval;
+int lvp, first;
 
-char buf[256];
+char buf[256], val[256];
 double tab[MAXROWS][MAXCOLS];
 double total[2][MAXROWS];
 double grantotal;
 char list[2][MAXROWS][WORDLEN];
-
-int stat;
-int align;
-double adjx, adjy;
-
-int j, m, fld, nlist[3];
-int select[3], ndim;
-int order[2][MAXROWS], valuesgiven[2], nval, dopercents;
+int stat, j, nlist[3], select[3], ndim;
+int order[2][MAXROWS], valuesgiven[2], dopercents;
 int forcevertical, forcehorizontal;
-/* int outputbyfreq[2], reverse[2]; */
 char ordering[2];
 char tmp[WORDLEN];
 char *GL_getok();
-double atof(), f;
+double atof();
 /* --- */
-char valuelist[2][256];
-int field[2];
-int ix;
-double tmpd;
-char savetable[80];
-int len;
-int irow;
-double gran[2];
-int davail;
-int curd;
-char numbuf[80];
-char selectex[80];
-int result;
-int doranges[2];
-char axis[2];
-int ixx;
-char hival[80], lowval[80];
-double hiv[MAXBINS], lowv[MAXBINS];
-double fval;
-char rangesepchar[5];
-char showrange[20];
-FILE *sfp;
-char rangespec[2][80];
+char *valuelist[2], *rangespec[2], *selectex, *rangesepchar, *showrange, *fieldnamelist, *numfmt;
+char axis[2], numbuf[80], hival[80], lowval[80], tag[80];
+int field[2], accumfield;
 int axisset[2]; /* added 1/11/00 scg */
-char tag[80];   /* added 1/11/00 scg */
-char fieldnamelist[256]; /* added 1/21/00 scg */
-int accumfield;
-double inc;
+int i, ix, showresults, irow, result, doranges[2], ixx;
+double gran[2], hiv[MAXBINS], lowv[MAXBINS], fval, inc;
 
 
 TDH_errprog( "pl proc tabulate" );
@@ -90,88 +63,58 @@ valuesgiven[0] = valuesgiven[1] = 0;
 dopercents = 0;
 forcevertical = 1; forcehorizontal = 0;
 ordering[0] = '?'; ordering[1] = '?';
-strcpy( savetable, "" );
 gran[0] = gran[1] = 0.0;
-strcpy( selectex, "" );
 doranges[0] = doranges[1] = 0;
 axis[0] = 'x';
 axis[1] = 'y';
-strcpy( rangesepchar, "-" );
-strcpy( showrange, "" );
-strcpy( rangespec[0], "" );
-strcpy( rangespec[1], "" );
-axisset[0] = axisset[1] = 0; /* was axis1 or axis2 set explicitly - added 1/11/00 scg */
+rangespec[0] = rangespec[1] = "";
+axisset[0] = axisset[1] = 0; 
 accumfield = -1;
-
+showresults = 0;
+selectex = "";
+rangesepchar = "-";
+showrange = "";
+numfmt = "%g";
+valuelist[0] = valuelist[1] = "";
+fieldnamelist = "";
 
 /* get attributes.. */
 first = 1;
 while( 1 ) {
-	line = getnextattr( first, attr, val, &lvp, &nt );
+	line = getnextattr( first, attr, &lvp );  
 	if( line == NULL ) break;
 	first = 0;
 	lineval = &line[lvp];
 
-	if( stricmp( attr, "datafield1" )==0 ) {
-		field[0] = fref( val ) - 1;
-		ndim = 1;
-		}
-	else if( stricmp( attr, "datafield2" )==0 ) {
-		field[1] = fref( val ) - 1;
-		ndim = 2;
-		}
-	else if( stricmp( attr, "accumfield" )==0 ) {
-		accumfield = fref( val ) -1;
-		}
-	else if( stricmp( attr, "valuelist1" )==0 ) strcpy( valuelist[0], lineval );
-	else if( stricmp( attr, "valuelist2" )==0 ) strcpy( valuelist[1], lineval );
-	else if( stricmp( attr, "doranges1" )==0 ) {
-		if( strnicmp( val, YESANS, 1 )==0 ) doranges[0] = 1;
-		else doranges[0] = 0;
-		}
-	else if( stricmp( attr, "doranges2" )==0 ) {
-		if( strnicmp( val, YESANS, 1 )==0 ) doranges[1] = 1;
-		else doranges[1] = 0;
-		}
-	else if( stricmp( attr, "showrangelowonly" )==0 ) { 
-		if( strnicmp( val, YESANS, 1 )==0 ) strcpy( showrange, "low" );
-		}
-	else if( stricmp( attr, "showrange" )==0 ) strcpy( showrange, val );
-	else if( stricmp( attr, "rangesepchar" )==0 ) strcpy( rangesepchar, val );
-	else if( stricmp( attr, "order1" )==0 ) {
-		if( strnicmp( val, "mag", 3 )==0 ) ordering[0] = 'm';
-		else if( strnicmp( val, "rev", 3 )==0 ) ordering[0] = 'r';
-		else if( strnicmp( val, "nat", 3 )==0 ) ordering[0] = 'n';
-		else if( stricmp( val, "none" )==0 ) ordering[0] = '0';
-		}
-	else if( stricmp( attr, "ordering2" )==0 ) {
-		if( strnicmp( val, "mag", 3 ) == 0 ) ordering[1] = 'm';
-		else if( strnicmp( val, "rev", 3 ) == 0 ) ordering[1] = 'r';
-		else if( strnicmp( val, "nat", 3 ) == 0 ) ordering[1] = 'n';
-		else if( stricmp( val, "none" ) == 0 ) ordering[1] = '0';
-		}
 
-	else if( stricmp( attr, "percents" )==0 ) {
-		if( strnicmp( val, YESANS, 1 )==0 ) dopercents = 1;
-		else dopercents = 0;
+	if( strcmp( attr, "datafield1" )==0 ) { field[0] = fref( lineval ) - 1; ndim = 1; }
+	else if( strcmp( attr, "datafield2" )==0 ) { field[1] = fref( lineval ) - 1; ndim = 2; }
+	else if( strcmp( attr, "accumfield" )==0 ) accumfield = fref( lineval ) -1;
+	else if( strcmp( attr, "valuelist1" )==0 ) valuelist[0] = lineval;
+	else if( strcmp( attr, "valuelist2" )==0 ) valuelist[1] = lineval;
+	else if( strcmp( attr, "doranges1" )==0 ) doranges[0] = getyn( lineval );
+	else if( strcmp( attr, "doranges2" )==0 ) doranges[1] = getyn( lineval );
+	else if( strcmp( attr, "showrangelowonly" )==0 ) { if( getyn( lineval ) ) showrange = "low"; }
+	else if( strcmp( attr, "showrange" )==0 ) showrange = lineval;
+	else if( strcmp( attr, "rangesepchar" )==0 ) rangesepchar = lineval;
+	else if( strcmp( attr, "resultfieldnames" )==0 ) fieldnamelist = lineval;
+	else if( strcmp( attr, "order1" )==0 ) {  /* mag, rev, nat, none */
+		if( lineval[0] == 'm' || lineval[0] == 'r' || lineval[1] == 'a' ) ordering[0] = lineval[0];
+		else ordering[0] = 0;
 		}
-	else if( stricmp( attr, "savetable" )==0 ) strcpy( savetable, val );
-
-	else if( stricmp( attr, "select" )==0 ) strcpy( selectex, lineval );
-
-	else if( stricmp( attr, "axis1" )==0 ) {
-		axis[0] = tolower(val[0]);
-		axisset[0] = 1;
+	else if( strcmp( attr, "order2" )==0 ) {
+		if( lineval[0] == 'm' || lineval[0] == 'r' || lineval[1] == 'a' ) ordering[1] = lineval[0];
+		else ordering[1] = 0;
 		}
-	else if( stricmp( attr, "axis2" )==0 ) {
-		axis[1] = tolower(val[0]);
-		axisset[1] = 1;
-		}
-
-	else if( stricmp( attr, "rangespec1" )==0 ) strcpy( rangespec[0], lineval );
-	else if( stricmp( attr, "rangespec2" )==0 ) strcpy( rangespec[1], lineval );
-
-	else Eerr( 1, "tabulate attribute not recognized", attr );
+	else if( strcmp( attr, "percents" )==0 ) dopercents = getyn( lineval );
+	else if( strcmp( attr, "showresults" )==0 || strcmp( attr, "savetable" )==0 ) showresults = getyn( lineval );
+	else if( strcmp( attr, "select" )==0 ) selectex = lineval;
+	else if( strcmp( attr, "axis1" )==0 ) { axis[0] = lineval[0]; axisset[0] = 1; }
+	else if( strcmp( attr, "axis2" )==0 ) { axis[1] = val[0]; axisset[1] = 1; }
+	else if( strcmp( attr, "rangespec1" )==0 ) rangespec[0] = lineval;
+	else if( strcmp( attr, "rangespec2" )==0 ) rangespec[1] = lineval;
+	else if( strcmp( attr, "numfmt" )==0 ) numfmt = lineval;
+	else Eerr( 1, "attribute not recognized", attr );
 	}
 
 
@@ -193,14 +136,6 @@ if( ordering[1] == '?' ) {
 /* now do the computation work.. */
 /* -------------------------- */
 
-if( savetable[0] != '\0' ) {
-	if( stricmp( savetable, "stdout" )==0 ) sfp = stdout;
-	else if( stricmp( savetable, "stderr" )==0 ) sfp = stderr;
-	else sfp = fopen( savetable, "w" );  /* diagnostics */
-	if( sfp == NULL ) sfp = Diagfp;
-	}
-
-
 /* initialize tables */
 grantotal = 0.0;
 for( i = 0; i < MAXROWS; i++ ) {
@@ -209,9 +144,12 @@ for( i = 0; i < MAXROWS; i++ ) {
 	order[0][i] = order[1][i] = i;
         }
 
+
 /* get value list (a comma or space delimited list of values; output distribution will
    be for only these values (in the order given?) */
 for( j = 0; j < ndim; j++ ) { /* for all dimensions (1 or 2).. */
+
+
 	if( valuelist[j][0] != '\0' ) {
 	    int ix;
 	    for( i = 0, ix = 0;  ; i++ ) {
@@ -223,7 +161,7 @@ for( j = 0; j < ndim; j++ ) { /* for all dimensions (1 or 2).. */
 			ixx = 0;
 			GL_getseg( lowval, tmp, &ixx, rangesepchar );
 			strcpy( hival, &tmp[ixx] );
-			if( stricmp( lowval, "c" )==0 && i > 0 ) lowv[i] = hiv[i-1]; /* contiguous*/
+			if( strcmp( lowval, "c" )==0 && i > 0 ) lowv[i] = hiv[i-1]; /* contiguous*/
 			else lowv[i] = Econv( axis[j], lowval );
 			if( Econv_error( ) ) {   /* a non-conformant value.. */
 				lowv[i] = PLHUGE; /* so that we know to compare 
@@ -232,9 +170,9 @@ for( j = 0; j < ndim; j++ ) { /* for all dimensions (1 or 2).. */
 				}
 			else 	{
 				hiv[i] = Econv( axis[j], hival );
-				Euprint( lowval, axis[j], lowv[i], "" );
-				if( tolower(showrange[0]) == 'l' ) strcpy( list[j][i], lowval );
-				else if( tolower(showrange[0]) == 'a' ) Euprint( list[j][i], axis[j], (lowv[i]+hiv[i])/2.0, "" );
+				Euprint( lowval, axis[j], lowv[i], numfmt );  /* lowval[80] */
+				if( showrange[0] == 'l' ) strcpy( list[j][i], lowval );
+				else if( showrange[0] == 'a' ) Euprint( list[j][i], axis[j], (lowv[i]+hiv[i])/2.0, numfmt ); /* [40] */
 
 				else sprintf( list[j][i], "%s%s%s", lowval, rangesepchar, hival );
 				}
@@ -246,6 +184,7 @@ for( j = 0; j < ndim; j++ ) { /* for all dimensions (1 or 2).. */
 	/* automatic bins */
 	if( rangespec[j][0] != '\0' ) {
 		double binsiz, hilimit, rw;
+		int nt;
 		doranges[j] = 1; /* implied */
 		nt = sscanf( rangespec[j], "%s %lf %s", lowval, &binsiz, hival );
 		if( nt < 2 || nt > 3 ) return( Eerr( 2740, "2 or 3 values expected in rangespec", 
@@ -264,10 +203,10 @@ for( j = 0; j < ndim; j++ ) { /* for all dimensions (1 or 2).. */
 			lowv[i] = rw;
 			rw += binsiz;
 			hiv[i] = rw;
-			Euprint( lowval, axis[j], lowv[i], "" );
-			Euprint( hival, axis[j], hiv[i], "" );
-			if( tolower(showrange[0]) == 'l' ) strcpy( list[j][i], lowval );
-			else if( tolower(showrange[0]) == 'a' ) Euprint( list[j][i], axis[j], (lowv[i]+hiv[i])/2.0, "" );
+			Euprint( lowval, axis[j], lowv[i], "" ); /* lowval[80] */
+			Euprint( hival, axis[j], hiv[i], "" ); /* hival[80] */
+			if( showrange[0] == 'l' ) strcpy( list[j][i], lowval );
+			else if( showrange[0] == 'a' ) Euprint( list[j][i], axis[j], (lowv[i]+hiv[i])/2.0, numfmt );  /* [40] */
 			else sprintf( list[j][i], "%s%s%s", lowval, rangesepchar, hival );
 			nlist[j]++;
 			if( rw > hilimit ) break;
@@ -279,22 +218,16 @@ for( j = 0; j < ndim; j++ ) { /* for all dimensions (1 or 2).. */
 		return( Eerr( 2052, "A values list must be given when doing ranges.", "" ));
 	}
 
-/* make a description string */
-if( ndim == 1 ) sprintf( buf, "// 1-way distribution on field %d %s", field[0]+1, selectex ); 
-if( ndim == 2 ) sprintf( buf, "// 2-way distribution on field %d (down) by field %d (across) %s", 
-	field[0]+1, field[1]+1, selectex ); 
-/* set a DMS variable to the description? */
-if( savetable[0] != '\0' ) fprintf( sfp, "%s\n", buf );
+
 
 
 /* process from data already read in earlier.. */
-Dsel = 0;
+if( Nrecords < 1 ) return( Eerr( 32, "No data has been read yet.", "" ) );
 
-if( Nrecords[Dsel] < 1 ) return( Eerr( 32, "No data has been read yet.", "" ) );
 
 /* process input data.. */
 ix = 0;
-for( irow = 0; irow < Nrecords[Dsel]; irow++ ) {
+for( irow = 0; irow < Nrecords; irow++ ) {
 
  	if( selectex[0] != '\0' ) { /* process against selection condition if any.. */
                 stat = do_select( selectex, irow, &result );
@@ -328,7 +261,7 @@ for( irow = 0; irow < Nrecords[Dsel]; irow++ ) {
 				/* otherwise, add the internal representation to list */
 				if( j == nlist[i] ) {
 					if( (i == 0 && nlist[i] >= MAXROWS) || (i == 1 && nlist[i] >= MAXCOLS) ) 
-					    fprintf( Errfp, "tabulate warning, sorry, table capacity exceeded, skipping %s\n", val );
+					    fprintf( PLS.errfp, "tabulate warning, sorry, table capacity exceeded, skipping %s\n", val );
 					else	{
 						if( !Econv_error()) 
 							sprintf( list[i][j], "%f", fval );
@@ -353,7 +286,7 @@ for( irow = 0; irow < Nrecords[Dsel]; irow++ ) {
 	                        /* add it to list */
 	                        if( j == nlist[i] ) { 
 					if( (i == 0 && nlist[i] >= MAXROWS) || (i == 1 && nlist[i] >= MAXCOLS) ) 
-						fprintf( Errfp, "tabulate warning, sorry, table capacity exceeded, skipping %s\n", val );
+						fprintf( PLS.errfp, "tabulate warning, sorry, table capacity exceeded, skipping %s\n", val );
 					else	{
 						strcpy( list[i][j], val ); 
 						select[i] = j; 
@@ -368,6 +301,7 @@ for( irow = 0; irow < Nrecords[Dsel]; irow++ ) {
                 else if( doranges[i] ) {
 			fval = Econv( axis[i], val );
 
+
 			/* go through list backwards so that boundary values are 
 			   put into higher category.. */
 			for( j = nlist[i] - 1; j >= 0; j-- ) {
@@ -429,64 +363,44 @@ else if( ndim == 2 ) {
 /* generate the results */
 /* --------------------- */
 
-Dsel = 1;
-
-davail = Dsize;
-ND[Dsel] = 0;
-StartD[Dsel] = ND[0];
-curd = StartD[Dsel];
+/* make a description string */
+if( showresults ) {
+        if( ndim == 1 ) fprintf( PLS.diagfp, "// proc tabulate has computed this distribution on field %d %s\n", field[0]+1, selectex );
+        else if( ndim == 2 ) fprintf( PLS.diagfp, "// proc tabulate has computed this 2-way distribution\n// on field %d (down) by field %d (across) %s\n",
+                field[0]+1, field[1]+1, selectex );
+	}
 
 
+
+/* PL_newdataset(); */
+PL_begindataset();
+
+ 
 /* ------------------ */
 /* for 1-way tables.. */
 /* ------------------ */
 if( ndim == 1 ) {
 
-	/* set PL data array dimensions for slot 2 */
-	Nfields[Dsel] = 2;
-	Nrecords[Dsel] = nlist[0];
-
-        sprintf( buf, "//%-20s    N", "Value" );
-	if( dopercents ) {
-		sprintf( buf, "%s  Pct", buf );
-		Nfields[Dsel] = 3;
-		}
-	if( savetable[0] != '\0' ) fprintf( sfp, "%s\n\n", buf );
-
 
 	/* do output lines.. */
         for( i = 0; i < nlist[0]; i++ ) {
-		if( !doranges[0] && axisset[0] ) 
-			Euprint( tag, axis[0], atof(list[0][order[0][i]]), "" );
+		if( !doranges[0] && axisset[0] ) Euprint( tag, axis[0], atof(list[0][order[0][i]]), "" ); /* tag[80] */
 		else strcpy( tag, list[0][order[0][i]] );
 
-		sprintf( buf, "%-20s  %5.0f", tag, tab[0][order[0][i]] );
-		if( dopercents ) sprintf( buf, "%s  %3.0f", buf, 
-			(tab[0][order[0][i]])/(total[0][0]+0.0001)*100  );
-		if( savetable[0] != '\0' ) fprintf( sfp, " %s\n", buf );
-		
-		/* put the results into the PL data array.. */
 
-		/* label */
-		catitem( tag, &curd, &davail );
+		PL_startdatarow();
+		PL_catitem( tag ); /* label */
 
 		/* n */
-		sprintf( buf, "%g", tab[0][order[0][i]] ); 
-		catitem( buf, &curd, &davail );
+		sprintf( buf, numfmt, tab[0][order[0][i]] ); 
+		PL_catitem( buf );
 
 		if( dopercents ) {
-			sprintf( buf, "%g", (double)(tab[0][order[0][i]])/(total[0][0]+0.0001)*100  );
-			catitem( buf, &curd, &davail );
+			sprintf( buf, numfmt, (double)(tab[0][order[0][i]])/(total[0][0]+0.0001)*100  );
+			PL_catitem( buf );
 			}
+		PL_enddatarow();
 		}
-	ND[Dsel] = curd - StartD[Dsel];
-
-	/* lastly, do total line */
-	sprintf( buf, "%-20s  %5.0f", "//Total:", total[0][0] );
-	if( dopercents ) sprintf( buf, "%s  100", buf );
-
-	/* Use DMS variable here? */
-	if( savetable[0] != '\0' ) fprintf( sfp, " %s\n", buf );
 	}
 
 
@@ -495,109 +409,54 @@ if( ndim == 1 ) {
 /* ------------------ */
 else if( ndim == 2 ) { 
 
-	/* set PL data array dimensions for slot 2 */
-	if( dopercents ) Nfields[Dsel] = ((nlist[1]+1)*2) + 1;
-	else Nfields[Dsel] = (nlist[1]+1) + 1; /* label, cols, rowtotal */
-	Nrecords[Dsel] = nlist[0];
-
-
-	/* build header */
-	sprintf( buf, "//%-20s", "" );
-	strcpy( fieldnamelist, "rowname" );
-        for( i = 0; i < nlist[1]; i++ ) {
-		if( !doranges[1] && axisset[1] ) 
-			Euprint( tag, axis[1], atof(list[1][order[1][i]]), "" );
-		else strcpy( tag, list[1][order[1][i]] );
-
-		sprintf( buf, "%s%6s  ", buf, tag );
-	 	if( dopercents ) sprintf( buf, "%s%s", buf, " Pct    " );
-
-		/* set up the column tags as data field names.. */
-		strcat( fieldnamelist, "," );
-		strcat( fieldnamelist, tag );
-		}
-
 	if( fieldnamelist[0] != '\0' ) {
 		definefieldnames( fieldnamelist );
-		fprintf( Diagfp, "proc tabulate: field names are now: %s\n", fieldnamelist );
+		fprintf( PLS.diagfp, "proc tabulate: field names are now: %s\n", fieldnamelist );
 		}
 	 
-	sprintf( buf, "%s%s  ", buf, "  Total" );
-	if( dopercents ) sprintf( buf, "%s%s", buf, " Pct    " );
-	if( savetable[0] != '\0' ) fprintf( sfp, "%s\n\n", buf );
-
 	/* do output lines.. */
         for( j = 0; j < nlist[0]; j++ ) {
-		if( !doranges[0] && axisset[0] ) 
-			Euprint( tag, axis[0], atof(list[0][order[0][j]]), "" );
+		if( !doranges[0] && axisset[0] ) Euprint( tag, axis[0], atof(list[0][order[0][j]]), "" ); /* tag[80] */
 		else strcpy( tag, list[0][order[0][j]] );
 
-		sprintf( buf, "%-20s", tag ); 
-
-		/* put label into PL data */
-		catitem( tag, &curd, &davail );
+		PL_startdatarow();
+		PL_catitem( tag ); /* label */
 
                 for( i = 0; i < nlist[1]; i++ ) {
-			sprintf( buf, "%s %5.0f  ", buf, tab[order[0][j]][order[1][i]] ); 
 
 			/* n */
-			sprintf( numbuf, "%g", tab[order[0][j]] [order[1][i]] );
-			catitem( numbuf, &curd, &davail );
+			sprintf( numbuf, numfmt, tab[order[0][j]] [order[1][i]] );
+			PL_catitem( numbuf );
 
 			if( dopercents ) {
-				sprintf( buf, "%s %3.0f    ", buf, 
-				  (double)(tab[order[0][j]][order[1][i]])/
-					(total[1][order[1][i]]+0.00001) * 100 ); 
-
-				sprintf( numbuf, "%g", 
-				  (double)(tab[order[0][j]][order[1][i]])/
-					(total[1][order[1][i]]+0.00001) * 100 ); 
-				catitem( numbuf, &curd, &davail );
+				sprintf( numbuf, numfmt, 
+				  	(double)(tab[order[0][j]][order[1][i]])/ (total[1][order[1][i]]+0.00001) * 100 ); 
+				PL_catitem( numbuf );
 				}
 			}
 		
-                sprintf( buf, "%s%5.0f  ", buf, total[0][order[0][j]] );
 		/* row total */
-		sprintf( numbuf, "%g", total[0][order[0][j]] );
-		catitem( numbuf, &curd, &davail );
+		sprintf( numbuf, numfmt, total[0][order[0][j]] );
+		PL_catitem( numbuf );
 
 		if( dopercents ) {
-			sprintf( buf, "%s   %3.0f", buf, 
-			  (double)(total[0][order[0][j]]) / (grantotal+0.00001) * 100 ); 
-			sprintf( numbuf, "%g", 
-			  (double)(total[0][order[0][j]]) / (grantotal+0.00001) * 100 ); 
-			catitem( numbuf, &curd, &davail );
+			sprintf( numbuf, numfmt, (double)(total[0][order[0][j]]) / (grantotal+0.00001) * 100 ); 
+			PL_catitem( numbuf );
 			}
 		
-		if( savetable[0] != '\0' ) fprintf( sfp, " %s\n", buf );
+		PL_enddatarow();
                 }
-	ND[Dsel] = curd - StartD[Dsel];
 
-	/* lastly, do total line.. */
-	sprintf( buf, "%-19s", "//Total:" );
-        for( i = 0; i < nlist[1]; i++ ) {
-		sprintf( buf, "%s  %5.0f ", buf, total[1][order[1][i]] );
-		if( dopercents ) sprintf( buf, "%s  100   ", buf ); 
-		}
-        sprintf( buf, "%s %5.0f", buf, grantotal );
-	if( dopercents ) sprintf( buf, "%s     100", buf );
-	/* set DMS var here?? */
-	if( savetable[0] != '\0' ) fprintf( sfp, " %s\n", buf );
-        }
+	}
 
 
-/* fprintf( Diagfp, "After Xtab:     ND=%d Nrecords=%d Nfields=%d\n", 
-	ND[Dsel], Nrecords[Dsel], Nfields[Dsel] ); */
+PL_finishdataset( 0, 0 );
 
-setintvar( "NRECORDS", Nrecords[Dsel] );
-setintvar( "NFIELDS", Nfields[Dsel] );
+if( showresults ) for( i = 0; i < Nrecords; i++ ) {
+	for( j = 0; j < Nfields; j++ ) fprintf( PLS.diagfp, "[%s]", da( i, j ) );
+	fprintf( PLS.diagfp, "\n" );
+	}
  
-/* for( i = 0; i < Nrecords[Dsel]; i++ ) {
- *      for( j = 0; j < Nfields[Dsel]; j++ ) fprintf( Diagfp, "|%s| ", da(i,j) );
- *      fprintf( Diagfp, "\n" );
- *      }
- */
-
 
 return( 0 );
 }
@@ -638,6 +497,8 @@ for( i = 0; i < nd; i++ ) {
         order[i] = mincell;
         used[mincell] = 1;
         }
+
+return( 0 );
 }
 
 /* ================ */
@@ -666,4 +527,11 @@ for( i = 0; i < nd; i++ ) {
         order[i] = firstcell;
         used[firstcell] = 1;
         }
+return( 0 );
 }
+
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
