@@ -1,11 +1,11 @@
-/* ploticus data display engine.  Software, documentation, and examples.  
- * Copyright 1998-2002 Stephen C. Grubb  (scg@jax.org).
- * Covered by GPL; see the file ./Copyright for details. */
-
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
 
 /* ROUTINES FOR WORKING WITH VARIOUS UNITS */
-/* These routines may be thought of as "sitting on top of" 
-   the routines in ./src/graphic.c 
+/* These routines may be thought of as "sitting on top of" the routines in ./src/graphic.c 
 
    References dmslib/dates.c and dmslib/times.c
 */
@@ -21,21 +21,37 @@
 
 static char unitdesc[2][20] = { "", "" };
 static int unittyp[2] = { 0, 0 };
-
 static int conv_errflag = 0;
-
 static int dashindate = 0;	 /* 1 if current date format contains dash(s), 0 otherwise */
 static char dashindateaxis = '0';
+static double catslide[2] = { 0.0, 0.0 };  /* for displaying clusters over categories */
 
 static int pex();
 static int do_pex();
-
-static double catslide[2] = { 0.0, 0.0 };  /* for displaying clusters over categories */
+static int setdatesub();
+static int convdatesub();
+static int evalbound();
 
 
 /* ============================ */
-/* ESETUNITS - set up for special units on an axis */
-Esetunits( axis, s )
+int
+PL_units_initstatic()
+{
+strcpy( unitdesc[0], "" ); strcpy( unitdesc[1], "" );
+unittyp[0] = 0; unittyp[1] = 0;
+conv_errflag = 0;
+dashindate = 0;
+dashindateaxis = '\0';
+catslide[0] = 0.0; catslide[1] = 0.0; 
+
+return( 0 );
+}
+
+
+/* ============================ */
+/* SETUNITS - set up for special units on an axis */
+int
+PL_setunits( axis, s )
 char axis;
 char *s;
 {
@@ -50,12 +66,14 @@ else i = 1;
 
 strcpy( unitdesc[i], "" );
 
+strcpy( tok, "" );
+strcpy( tok2, "" );
 nt = sscanf( s, "%s %s", tok, tok2 );
 
 
-if( stricmp( tok, "linear" )==0 ) unittyp[i] = LINEAR;
+if( strcmp( tok, "linear" )==0 ) unittyp[i] = LINEAR;
 
-else if( stricmp( tok, "date" )==0 ) {
+else if( strcmp( tok, "date" )==0 ) {
 	unittyp[i] = DATE;
 	dashindate = 0; dashindateaxis = '0';
 	if( nt == 2 ) {
@@ -71,7 +89,7 @@ else if( stricmp( tok, "date" )==0 ) {
 		}
 	}
 
-else if( stricmp( tok, "time" )==0 ) {
+else if( strcmp( tok, "time" )==0 ) {
 	unittyp[i] = TIME;
 	if( nt == 2 ) {
 		stat = DT_settimefmt( tok2 );
@@ -79,20 +97,20 @@ else if( stricmp( tok, "time" )==0 ) {
 		}
 	}
 
-else if( strnicmp( tok, "datetime", 8 )==0 ) {
+else if( strncmp( tok, "datetime", 8 )==0 ) {
 	unittyp[i] = DATETIME;
 	dashindate = 0; dashindateaxis = '0';
-	if( nt == 2 ) {
-		stat = DT_setdatetimefmt( tok2, &tok[8] );
-		if( stat != 0 ) return( -1 ) ;
-		}
+	/* if( nt == 2 ) { */ /* condition removed scg 9/29/03 */
+	stat = DT_setdatetimefmt( tok2, &tok[8] );
+	if( stat != 0 ) return( -1 ) ;
+	/* } */
 	DT_getdatefmt( tok2 );
 	/* check for - in date format- this messes up arithmetic */
 	for( j = 0, slen = strlen( tok2 ); j < slen; j++ ) {
 		if( tok2[j] == '-' ) { dashindate = 1; dashindateaxis = axis; break; }
 		}
 	}
-else if( stricmp( tok, "categories" )==0 ) {
+else if( strcmp( tok, "categories" )==0 ) {
 	unittyp[i] = CATEGORIES;
 	catslide[0] = 0.0; catslide[1] = 0.0;
 	}
@@ -100,8 +118,9 @@ else return( -1 ); /* unknown type */
 return( 0 );
 }
 /* ============================ */
-/* EGETUNITS - report on what units are in effect on the given axis. */
-Egetunits( axis, result )
+/* GETUNITS - report on what units are in effect on the given axis. */
+int
+PL_getunits( axis, result )
 char axis;
 char *result;
 {
@@ -119,8 +138,9 @@ return( 0 );
 }
 
 /* ============================ */
-/* EGETUNITSUBTYPE - report on subtype of units are in effect on the given axis. */
-Egetunitsubtype( axis, result )
+/* GETUNITSUBTYPE - report on subtype of units are in effect on the given axis. */
+int
+PL_getunitsubtype( axis, result )
 char axis;
 char *result;
 {
@@ -134,8 +154,9 @@ return( 0 );
 }
 
 /* ============================ */
-/* ESETSCALE - set scaling for an axis - special units supported */
-Esetscale( axis, alo, ahi, scalelo, scalehi )
+/* SETSCALE - set scaling for an axis - special units supported */
+int
+PL_setscale( axis, alo, ahi, scalelo, scalehi )
 char axis;
 double alo, ahi;
 char *scalelo, *scalehi;
@@ -159,15 +180,15 @@ else	{
 
 if( axis == 'x' ) return( Escale_x( alo, ahi, slo, shi ) );
 
-else if( axis == 'y' ) return( Escale_y( alo, ahi, slo, shi ) );
+else if( axis == 'y' ) return( Escale_y( alo, ahi, slo, shi ) ); 
 
 else return( Eerr( 209, "invalid axis", "" ));
 }
 
 /* ============================ */
-/* ECONV - return linear equivalent of special units */
+/* CONV - return linear equivalent of special units */
 double
-Econv( axis, s )
+PL_conv( axis, s )
 char axis;
 char *s;
 {
@@ -218,31 +239,14 @@ else if( unittyp[i] == DATETIME ) {
 	}
 
 else if( unittyp[i] == CATEGORIES ) { 
-	int j;
-	int axslot;
-	if( axis == 'x' ) axslot = 0;
-	else axslot = 1;
+	int catj;
 	/* look up the category.. */
-	for( j = 0; j < Ncats[ axslot ]; j++ ) {
-		if( Catcompmethod == 0 ) {   	  /* contains */
-			if( strnicmp( s, Cats[axslot][j], strlen(s) )==0 ) break;
-			}
-		else if( Catcompmethod == -1 ) {  /* exact */
-			if( stricmp( s, Cats[axslot][j] )==0 ) break;
-			}
-		else if( Catcompmethod > 0 ) {   /* specified length */
-			if( strnicmp( s, Cats[axslot][j], Catcompmethod )==0 ) break;
-			}
-		else if( Catcompmethod < -10 ) {   /* specified length */
-			if( GL_fuzzymatch( Cats[axslot][j], s, strlen(s), (15+Catcompmethod) ) ) break;
-			}
-		}
-	if( j >= Ncats[ axslot ] ) {  /* not found */
-		conv_errflag = 1; 
-		return( 0.0 ); 
-		}
-	if( axis == 'x' ) f = (double)(j+1) + catslide[0]; /* +1 because we will never want origin */
-	else f = EDYhi-((double)(j+1) + catslide[1]);  /* cats always top down in Y */
+	catj = PL_findcat( axis, s );
+	if( catj < 0 ) { conv_errflag = 1; return( 0.0 ); }
+
+	if( axis == 'x' ) f = (double)( catj + 1 ) + catslide[0]; /* +1 because we will never want origin */
+	else f = EDYhi-((double)( catj + 1 ) + catslide[1]);   /* cats always top down in Y */  /* 9/28/07- it seems like it should be: -catslide */
+											       /* but changing this now could break lots of things*/
 	return( f );
 	}
 
@@ -252,15 +256,17 @@ else	{
 	}
 }
 /* ============================= */
-/* ECONV_ERROR check of Econv() error status */
-Econv_error()
+/* CONV_ERROR check of Econv() error status */
+int
+PL_conv_error()
 {
 return( conv_errflag );
 }
 
 /* ============================= */
-/* EUPRINT - produce a string containing the external representation of a value */
-Euprint( result, axis, f, format )
+/* UPRINT - produce a string containing the external representation of a value.  Result is copied into 'result'. */
+int
+PL_uprint( result, axis, f, format )
 char *result;
 char axis;
 double f;
@@ -273,28 +279,34 @@ int stat;
 if( axis == 'x' ) i = 0;
 else i = 1;
 
+strcpy( result, "???" );
+
 if( unittyp[i] == LINEAR ) {
 	/* when generating incremental axes moving from negative to positive, for zero sprintf sometimes 
 	   gives -0.00 or very tiny values like -5.5579e-17.  The following is a workaround.. scg 7/5/01 */
-	if( f < 0.0000000000001 && f > -0.0000000000001 ) f = 0.0;
+	/* if( f < 0.0000000000001 && f > -0.0000000000001 ) f = 0.0; */  /* moved this to proc_axis() scg 10/1/03 */
 
 	if( format[0] == '\0' ) sprintf( result, "%g", f );
+	else if( strncmp( format, "autoround", 9 )==0 ) {
+		if( format[9] == '\0' ) strcpy( result, GL_autoroundf( f, 0 ));
+		else strcpy( result, GL_autoroundf( f, atoi( &format[9] ) ));
+		}
 	else sprintf( result, format, f );
 
-	if( Bignumspacer ) rewritenums( result ); /* rewrite w/various spacing, decimal pt options*/
+	if( PLS.bignumspacer ) rewritenums( result ); /* rewrite w/various spacing, decimal pt options*/
 	}
 
 
 else if( unittyp[i] == DATE ) {
 	stat = DT_fromjul( (long) f, s );
 	if( stat != 0 ) {
-		fprintf( Errfp, "[invalid f value: %g stat=%d]", f, stat );
+		fprintf( PLS.errfp, "[invalid f value: %g stat=%d]", f, stat );
 		return( Eerr( 801, "error in parsing date", "" ));
 		}
 
 	if( format[0] != '\0' )  {
 		stat = DT_formatdate( s, format, result );
-		if( stat != 0 ) return( Eerr( 802, "error in converting date format", format ));
+		if( stat != 0 ) strcpy( result, s );  /* was an error   scg 12/18/03 */
 		}
 	else strcpy( result, s );
 	}
@@ -304,7 +316,7 @@ else if( unittyp[i] == TIME ) {
 	if( stat != 0 ) return( Eerr( 803, "error in parsing time", s));
 	if( format[0] != '\0' ) {
 		stat = DT_formattime( s, format, result );
-		if( stat != 0 ) return( Eerr( 804, "error in converting time format", format ));
+		if( stat != 0 ) strcpy( result, s );  /* was an 804 error  scg 12/18/03 */
 		}
 	else strcpy( result, s );
 	}
@@ -313,7 +325,7 @@ else if( unittyp[i] == DATETIME ) {
 	stat = DT_days2datetime( f, s );
 	if( format[0] != '\0' ) {
 		stat = DT_formatdatetime( s, format, result );
-		if( stat != 0 ) return( Eerr( 804, "error in converting datetime format", format ));
+		if( stat != 0 ) strcpy( result, s );  /* was an 804 error  scg 12/18/03 */
 		}
 	else strcpy( result, s );
 	}
@@ -326,8 +338,8 @@ return( 0 );
 }
 
 /* ========================= */
-/* EPOSEX - evaluate position expression and return absolute.
-   ELENEX - evaluate length expression and return absolute.  No special units
+/* POSEX - evaluate position expression and return absolute.
+   LENEX - evaluate length expression and return absolute.  No special units
 		are allowed in a length, thus for date scaling, value is in days;
 		for time scaling, value is in seconds.
 
@@ -344,7 +356,8 @@ return( 0 );
 
 	Returns 0 if ok, 1 on error.
 */
-Eposex( val, axis, result )
+int
+PL_posex( val, axis, result )
 char *val, axis;
 double *result;
 {
@@ -352,7 +365,8 @@ return( pex( val, axis, result, 0 ) );
 }
 
 /* ------------ */
-Elenex( val, axis, result )
+int
+PL_lenex( val, axis, result )
 char *val, axis;
 double *result;
 {
@@ -368,10 +382,7 @@ int mode; /* 0 = position,  1 = length */
 			need to normalize against minima */
 {
 int i;
-int len;
-double f;
 int stat;
-char modifier[12];
 char *subval[2];
 int nval;
 double result2;
@@ -382,6 +393,7 @@ int slen;
 
 if( in[0] == '\0' ) { *result = 0.0;  return( 1 ); }
 
+conv_errflag = 0; /* scg 4/22/02 */
 strcpy( val, in );
 nval = 1;
 op = 1;
@@ -391,7 +403,8 @@ subval[0] = &val[0];
 neg_offset_allowed = 1;
 if( dashindate && dashindateaxis == axis ) neg_offset_allowed = 0;
 for( i = 1, slen = strlen( val ); i < slen; i++ ) {
-	if( val[i] == '+' || (val[i] == '-' && neg_offset_allowed) ) {
+	/* the following condition was changed scg 10/1/03 to allow scientific notation values.. */
+	if( ( val[i] == '+' || (val[i] == '-' && neg_offset_allowed) ) && tolower( val[i-1] ) != 'e' ) {  /* EMB-MIN */
 		if( val[i] == '+' ) op = 1;
 		else if( val[i] == '-' ) op = 2;
 		val[i] = '\0';
@@ -417,18 +430,16 @@ char axis;
 double *result;
 int mode;	/* 0 = position  1 = length (no sp. units in lengths)  */
 {
-int i;
-int len;
 char buf[255];
-int stat;
-char modifier[12];
 double atof();
-int nt;
+int nt, k, found;
 double f;
+int len;
+char modifier[12];
 
 /* check for min and max.. */
-if( strnicmp( val, "min", 3 )==0 ) { *result = Elimit( axis, 'l', 'a' ); return( 0 ); }
-else if( strnicmp( val, "max", 3 )==0 ) { *result = Elimit( axis, 'h', 'a' ); return( 0 ); }
+if( strncmp( val, "min", 3 )==0 ) { *result = Elimit( axis, 'l', 'a' ); return( 0 ); }
+else if( strncmp( val, "max", 3 )==0 ) { *result = Elimit( axis, 'h', 'a' ); return( 0 ); }
 
 
 strcpy( buf, val );
@@ -448,7 +459,7 @@ if( nt < 1 ) strcpy( modifier, "(s)" ); /* assume scaled */
 else *result = f;
 
 /* scaled units */
-if( stricmp( modifier, "(s)" )==0 ) {
+if( strcmp( modifier, "(s)" )==0 ) {
 	if( !scalebeenset() )
 		return( Eerr( 51, "scaled units specified but y scaling not set", val ) );
 	if( mode == 1 ) {
@@ -457,13 +468,22 @@ if( stricmp( modifier, "(s)" )==0 ) {
 		}
 	if( mode == 0 ) {
 		*result = Ea( axis, Econv( axis, buf ) );
-		if( Econv_error() ) return( 1 );
+		if( Econv_error() ) {
+			/* try again, this time converting any underscores to spaces... 
+			 * allows catnames (etc) w/ embedded spaces to be usable for positioning.   added     scg 3/6/09 */
+			for( k = 0, found = 0; buf[k] != '\0'; k++ ) if( buf[k] == '_' ) { buf[k] = ' '; found = 1; }
+			if( found ) { /* try again only if there actually was an underscore in there.. */
+				*result = Ea( axis, Econv( axis, buf ) );
+				if( Econv_error() ) return( 1 );
+				}
+			else return( 0 );
+			}
 		else return( 0 );
 		}
 	}
 
 /* convert absolute centimeter specs to inches */
-if( Using_cm && (  modifier[0] == '\0' || stricmp( modifier, "(a)" )==0 ) ) {
+if( PLS.usingcm && (  modifier[0] == '\0' || strcmp( modifier, "(a)" )==0 ) ) {
 	*result = *result/ 2.54;
 	}
 
@@ -471,8 +491,9 @@ return( 0 );
 }
 
 /* ============================= */
-/* EU - take s in special units and return absolute */
-double Eu( axis, s )
+/* U - take s in special units and return absolute */
+double 
+PL_u( axis, s )
 char axis;
 char *s;
 {
@@ -489,6 +510,7 @@ return( Ea( axis, Econv( axis, s ) ) );
 
 	If date format includes embedded dashes then subtraction is disabled.
 */
+static int
 evalbound( val, axis, result )
 char *val;
 char axis;
@@ -499,8 +521,10 @@ int op, nval, i, slen;
 
 /* check for embedded + and -  (skip first character - could be unary)*/
 subval[0] = &val[0];
+nval = 0;
 for( i = 1, slen = strlen( val ); i < slen; i++ ) {
-        if( val[i] == '+' || (val[i] == '-' && !dashindate) ) {
+	/* the following condition was changed scg 10/1/03 to allow scientific notation values.. */
+        if( ( val[i] == '+' || (val[i] == '-' && !dashindate) ) && tolower(val[i-1]) != 'e' ) { 	/* EMB-MIN */
                 if( val[i] == '+' ) op = 1;
                 else if( val[i] == '-' ) op = 2;
                 val[i] = '\0';
@@ -522,24 +546,27 @@ return( 0 );
 /* date types that are special to ploticus
 	(generally yymm and Q type notations)
 */
+static int
 setdatesub( tok, desc )
 char *tok, *desc;
 {
-if( stricmp( tok, "yymm" )==0 ) { strcpy( desc, "yymm" ); sprintf( tok, "yymmdd" ); }
-else if( stricmp( tok, "yymmm" )==0 ) { strcpy( desc, "yymmm" ); sprintf( tok, "ddmmmyy" ); }
-else if( GL_slmember( tok, "yy?mm" )) { strcpy( desc, "yy/mm" ); sprintf( tok, "yy/mm/dd" ); }
-else if( GL_slmember( tok, "yyyy?mm" )) { strcpy( desc, "yyyy/mm" ); sprintf( tok, "yyyy/mm/dd" ); }
-else if( GL_slmember( tok, "yy" )) { strcpy( desc, "yy" ); sprintf( tok, "yymmdd" ); }
-else if( GL_slmember( tok, "mm?yy" )) { strcpy( desc, "mm/yy" ); sprintf( tok, "dd/mm/yy" ); }
-else if( GL_slmember( tok, "mm?yyyy" )) { strcpy( desc, "mm/yyyy" ); sprintf( tok, "dd/mm/yyyy" ); }
-else if( GL_slmember( tok, "yyqn" )) { strcpy( desc, "yyqn" ); sprintf( tok, "yy/mm/dd" ); }
-else if( GL_slmember( tok, "yyyyqn" )) { strcpy( desc, "yyyyqn" ); sprintf( tok, "yyyy/mm/dd" ); }
-else if( GL_slmember( tok, "nqyy" )) { strcpy( desc, "nqyy" ); sprintf( tok, "yy/mm/dd" ); }
-else if( GL_slmember( tok, "nqyyyy" )) { strcpy( desc, "nqyyyy" ); sprintf( tok, "yyyy/mm/dd" ); }
+if( strcmp( tok, "yymm" )==0 ) { strcpy( desc, "yymm" ); strcpy( tok, "yymmdd" ); }
+else if( strcmp( tok, "yymmm" )==0 ) { strcpy( desc, "yymmm" ); strcpy( tok, "ddmmmyy" ); }
+else if( GL_slmember( tok, "yy?mm" )) { strcpy( desc, "yy/mm" ); strcpy( tok, "yy/mm/dd" ); }
+else if( GL_slmember( tok, "yyyy?mm" )) { strcpy( desc, "yyyy/mm" ); strcpy( tok, "yyyy/mm/dd" ); }
+else if( GL_slmember( tok, "yy" )) { strcpy( desc, "yy" ); strcpy( tok, "yymmdd" ); }
+else if( GL_slmember( tok, "yyyy" )) { strcpy( desc, "yyyy" ); strcpy( tok, "yyyy/mm/dd" ); }  /* added scg 2/2/05 */
+else if( GL_slmember( tok, "mm?yy" )) { strcpy( desc, "mm/yy" ); strcpy( tok, "dd/mm/yy" ); }
+else if( GL_slmember( tok, "mm?yyyy" )) { strcpy( desc, "mm/yyyy" ); strcpy( tok, "dd/mm/yyyy" ); }
+else if( GL_slmember( tok, "yyqn" )) { strcpy( desc, "yyqn" ); strcpy( tok, "yy/mm/dd" ); }
+else if( GL_slmember( tok, "yyyyqn" )) { strcpy( desc, "yyyyqn" ); strcpy( tok, "yyyy/mm/dd" ); }
+else if( GL_slmember( tok, "nqyy" )) { strcpy( desc, "nqyy" ); strcpy( tok, "yy/mm/dd" ); }
+else if( GL_slmember( tok, "nqyyyy" )) { strcpy( desc, "nqyyyy" ); strcpy( tok, "yyyy/mm/dd" ); }
 return( 0 );
 }
 
 /* ----- */
+static int
 convdatesub( desc, s, result )
 char *desc, *s, *result;
 {
@@ -568,6 +595,9 @@ else if( strcmp( desc, "yyyy/mm" )==0 ) {
 else if( strcmp( desc, "yy" )==0 ) {
 	if( slen == 2 ) sprintf( result, "%s0101", s );  
 	}
+else if( strcmp( desc, "yyyy" )==0 ) {  /* added scg 2/2/05 */
+	if( slen == 4 ) sprintf( result, "%s/01/01", s );  
+	}
 else if( strcmp( desc, "mm/yy" )==0 ) {
 	if( slen == 5 ) sprintf( result, "01/%s", s );  
 	result[2] = '/'; result[5] = '/';
@@ -577,16 +607,16 @@ else if( strcmp( desc, "mm/yyyy" )==0 ) {
 	result[2] = '/'; result[5] = '/';
 	}
 else if( strcmp( desc, "yyqn" )==0 ) {
-	if( slen == 4 ) { sprintf( result, s ); quar = s[3]; placeconv = 2; }
+	if( slen == 4 ) { strcpy( result, s ); quar = s[3]; placeconv = 2; }
 	}
 else if( strcmp( desc, "yyyyqn" )==0 ) {
-	if( slen == 6 ) { sprintf( result, s ); quar = s[5]; placeconv = 4; }
+	if( slen == 6 ) { strcpy( result, s ); quar = s[5]; placeconv = 4; }
 	}
 else if( strcmp( desc, "nqyy" )==0 ) {
-	if( slen == 4 ) { sprintf( result, &s[2] ); quar = s[0]; placeconv = 2; }
+	if( slen == 4 ) { strcpy( result, &s[2] ); quar = s[0]; placeconv = 2; }
 	}
 else if( strcmp( desc, "nqyyyy" )==0 ) {
-	if( slen == 6 ) { sprintf( result, &s[2] ); quar = s[0]; placeconv = 4; }
+	if( slen == 6 ) { strcpy( result, &s[2] ); quar = s[0]; placeconv = 4; }
 	}
 else Eerr( 4802, "unrecognized subscale type", desc );
 
@@ -602,7 +632,8 @@ return( 0 );
 }
 
 /* ======================= */
-Esetcatslide( axis, amount )
+int
+PL_setcatslide( axis, amount )
 char axis;
 double amount;
 {
@@ -612,7 +643,8 @@ return( 0 );
 }
 /* ======================= */
 /* ES_INR - see if a string-based value is in range for the given axis */
-Es_inr( axis, val )
+int
+PL_s_inr( axis, val )
 char axis;
 char *val;
 {
@@ -627,7 +659,8 @@ else return( 0 );
 }
 /* ======================= */
 /* EF_INR - see if a float-based value is in range for the given axis */
-Ef_inr( axis, val )
+int
+PL_f_inr( axis, val )
 char axis;
 double val;
 {
@@ -641,3 +674,8 @@ else return( 0 );
 }
 
 
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
