@@ -1,13 +1,17 @@
-/* ploticus data display engine.  Software, documentation, and examples.  
- * Copyright 1998-2002 Stephen C. Grubb  (scg@jax.org).
- * Covered by GPL; see the file ./Copyright for details. */
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
 
+#include <stdio.h>
 
 static double xscale, yscale;
 static int window_height, window_width;
 
 /* ====================== set up user coords */
-Esetwinscale( width, height, x_max, y_max )
+int
+PLG_setwinscale( width, height, x_max, y_max )
 int width, height;
 double x_max, y_max;
 {
@@ -15,9 +19,11 @@ xscale = (double)(width) / (x_max);
 yscale = (double)(height) / (y_max);
 window_height = height;
 window_width = width;
+return( 0 );
 }
 /* ==================== scale in x to window size */
-Exsca( f )
+int
+PLG_xsca( f )
 double f;
 {
 int iout;
@@ -26,7 +32,8 @@ return( iout );
 }
 
 /* ==================== inverse of xsca */
-double Exsca_inv( i )
+double 
+PLG_xsca_inv( i )
 int i;
 {
 double out;
@@ -35,7 +42,8 @@ return( out );
 }
 
 /* ==================== scale in y to window size */
-Eysca( f )
+int
+PLG_ysca( f )
 double f;
 {
 int iout;
@@ -44,7 +52,8 @@ return( iout );
 }
 
 /* ===================== inverse of ysca */
-double Eysca_inv( i )
+double 
+PLG_ysca_inv( i )
 int i;
 {
 double out;
@@ -52,3 +61,8 @@ out = (double)(window_height-i) / yscale;
 return( out );
 }
 
+/* ======================================================= *
+ * Copyright 1998-2005 Stephen C. Grubb                    *
+ * http://ploticus.sourceforge.net                         *
+ * Covered by GPL; see the file ./Copyright for details.   *
+ * ======================================================= */
